"""Contains functions for checking the latest version of this package"""
import sys
import tempfile
import os
from time import sleep
from typing import Optional, Tuple

import requests
from packaging import version
from diskcache import Cache


ONE_DAY_IN_SECONDS = 60 * 60 * 24
cache = Cache(os.path.join(tempfile.gettempdir(), "terrawrap_version_cache"))


def version_check(current_version: str) -> bool:
    """
    Print a warning message if a stale version of Terrawrap is detected.
    :param current_version: The currently installed version of Terrawrap.
    :return: True if the version of Terrawrap is stale.
    """
    try:
        latest_stable, latest_rc = get_latest_version(current_version=current_version)
        current = version.parse(current_version)
        is_stale = False

        if version.parse(latest_stable) > current:
            print(
                "WARNING: Your version of Terrawrap is stale!",
                f"You have version '{current_version}' but the latest is '{latest_stable}'",
                "Please upgrade as soon as possible!\n pip install --upgrade terrawrap \n",
                sep="\n",
                file=sys.stderr,
            )
            is_stale = True

        if latest_rc and version.parse(latest_rc) > current:
            print(
                f"NOTE: A release candidate {latest_rc} is available.",
                f"\n pip install terrawrap=={latest_rc} \n",
                sep="",
                file=sys.stderr,
            )

        if is_stale:
            sleep(1)

        return is_stale
    except Exception as exp:
        print(
            f"WARNING: Encountered some error while checking for latest version of Terrawrap: {repr(exp)}",
        )
    return False


# We supply current_version so that the cache is invalidated when you install a new version of Terrawrap.
# pylint: disable=unused-argument
@cache.memoize(expire=ONE_DAY_IN_SECONDS)
def get_latest_version(current_version: str) -> Tuple[str, Optional[str]]:
    """
    Get the latest stable and RC versions of Terrawrap from PyPI.
    Caches this lookup for one day locally.
    :param current_version: The current version of Terrawrap (used for cache invalidation).
    :return: Tuple of (latest stable version, latest RC version or None).
    """
    response = requests.get(
        "https://pypi.python.org/pypi/terrawrap/json", timeout=5
    ).json()
    all_versions = [version.parse(v) for v in response["releases"]]
    stable_versions = [v for v in all_versions if not v.is_prerelease]
    rc_versions = [v for v in all_versions if v.is_prerelease]
    latest_stable = str(max(stable_versions))
    latest_rc = str(max(rc_versions)) if rc_versions else None
    return latest_stable, latest_rc



"""Module for containing convenience functions around path manipulation"""
import os
import re
import subprocess
from collections import defaultdict
from pathlib import Path
from typing import Dict, Optional, Set, Union

from networkx import DiGraph

GIT_REPO_REGEX = r"(URL)?.*/([\w-]*)(?:\.git)?"


def get_absolute_path(path: str, root_dir: Optional[str] = None) -> str:
    """
    Convenience function for determining the full path to a file or directory.
    A RuntimeError will be raised if the given path does not appear to point to either a file or a directory.
    :param path: The path. Can be either relative from the cwd or an absolute path.
    :param root_dir: The root directory to use instead of the cwd.
    :return: An absolute path for the given path.
    """
    if os.path.isabs(path):
        path = os.path.abspath(path)
    else:
        path = os.path.abspath(os.path.join(root_dir or os.getcwd(), path))

    return path


def get_symlinks(directory: str) -> Dict[str, Set[str]]:
    """
    Recursively walk a directory and return a dict of all symlinks
    :param directory:
    :return: dict of symlink source to set of paths that link to that source
    """
    links: Dict[str, Set[str]] = defaultdict(set)
    # pylint: disable=unused-variable
    for current_dir, dirs, files in os.walk(directory, followlinks=True):
        if ".terraform" in current_dir:
            continue

        if os.path.islink(current_dir):
            link_source = os.path.join(
                os.path.dirname(current_dir), os.readlink(current_dir)
            )
            links[os.path.normpath(link_source)].add(os.path.normpath(current_dir))

    return dict(links)


def get_file_graph(directory: str) -> DiGraph:
    """
    Recursively walk a directory and return a graph of all files and symlinks
    :param directory:
    :return: graph of file names to parent directories and symlink destinations
    """
    graph = DiGraph()
    for current_dir, dirs, files in os.walk(directory):
        if ".terraform" in current_dir or ".git" in current_dir:
            continue

        if current_dir not in graph.nodes:
            graph.add_node(current_dir)

        # for every file in a dir, create a node and an edge pointing to the parent dir
        # also create an edge for symlinks to symlink source
        for path in files:
            norm_path = os.path.normpath(os.path.join(current_dir, path))
            if norm_path not in graph.nodes:
                graph.add_node(norm_path)

            graph.add_edge(norm_path, current_dir)

            if os.path.islink(norm_path):
                link_source = os.path.normpath(
                    os.path.join(current_dir, os.readlink(norm_path))
                )

                if link_source not in graph.nodes:
                    graph.add_node(link_source)

                graph.add_edge(link_source, norm_path)

        # handle dirs the same way as files but don't create a node/edge for them unless they are a symlink
        for path in dirs:
            norm_path = os.path.normpath(os.path.join(current_dir, path))
            if os.path.islink(norm_path):
                link_source = os.path.normpath(
                    os.path.join(current_dir, os.readlink(norm_path))
                )

                if link_source not in graph.nodes:
                    graph.add_node(link_source)

                graph.add_edge(link_source, norm_path)
    return graph


def calc_repo_path(path: Union[str, Path]) -> str:
    """
    Convenience function for taking an absolute path to a TF directory and returning the path to that
    directory relative to the repo.
    :param path: The absolute path to a TF directory.
    :return: New path to the TF directory
    """
    path = str(path)

    byte_output = subprocess.check_output(
        ["git", "remote", "show", "origin", "-n"], cwd=path
    )
    output = byte_output.decode("utf-8", errors="replace")

    match = re.search(GIT_REPO_REGEX, output)
    if match:
        repo_name = match.group(2)
    else:
        raise RuntimeError("Could not determine git repo name, are we in a git repo?")

    return f'{repo_name}/{path[path.index("/config") + 1:]}'



# pylint:disable=C0114,C0301
import os
from functools import lru_cache
from os.path import relpath
from pathlib import Path
from typing import List, Union, Iterable, Any

import boto3
import git
from botocore.exceptions import ClientError
import hcl2

from terrawrap.utils.config import find_variable_files, parse_variable_files
from terrawrap.utils.path import calc_repo_path
from terrawrap.utils.terminal_colors import TerminalColors as Colors


class ConfigMover:
    """
    Class for moving terraform configuration between directories.
    This is done by moving .tfstate file in appropriate S3 bucket.
    """

    tf_state_file_extension = ".tfstate"

    def __init__(
        self,
        source_directory: Path,
        target_directory: Path,
    ):
        self.source_directory = source_directory
        self.target_directory = target_directory
        self.source_directory_abs = self.source_directory.absolute()
        self.target_directory_abs = self.target_directory.absolute()
        # we're using current path instead of the above in case they don't exist
        self._repo = git.Repo(os.getcwd(), search_parent_directories=True)
        self._s3 = boto3.client("s3")

    @property
    def s3_client(self):
        """S3 client"""
        return self._s3

    @property
    def repo(self) -> git.Repo:
        """Git repository"""
        return self._repo

    @lru_cache
    def _build_state_file_s3_key(self, terraform_path: Union[str, Path]) -> str:
        """
        Given path (absolute or relative) of terraform directory,
        return name of this directory's state file, example:
        Documents/terraform-config/config/aws/app/ ->
        terraform-config/config/aws/app.tfstate
        """
        return f"{calc_repo_path(terraform_path)}{self.tf_state_file_extension}"

    @lru_cache
    def _find_auto_variable(
        self, path: Path, variable_name: str, strict: bool = True
    ) -> Any:
        variables = parse_variable_files(self._find_variable_files(path))
        if strict:
            return variables[
                variable_name
            ]  # just in case variable_name is actually present, but is None
        return variables.get(variable_name)

    @lru_cache
    def _find_variable_files(self, path: Path) -> List[str]:
        """Find and cache .auto.tfvars files available at the scope of given path"""
        return find_variable_files(str(path))

    def _adjust_modules_sources(self, tf_files: Iterable[Path]):
        """
        For each module block in tf_files, adjusts path to its source depending on the location of target directory,
        for example, given that the target directory is one level deeper than the source directory:
        source = "../../../../../../modules/aws/s3_bucket/v3" ->
        source = "../../../../../../../modules/aws/s3_bucket/v3"
        This method works with modules paths virtually - it doesn't check their existence nor looks up them whatsoever.
        """

        for path in tf_files:
            content = path.read_text()
            config = hcl2.loads(content)
            modules = config.get("module", [])

            for module in modules:
                # get modules source relative path, e.g. '../../../modules/aws/s3_bucket/v3'
                module_source = list(module.values())[0]["source"]
                # convert above to absolute path
                module_source_absolute = (path.parent / module_source).resolve()
                # find modules source path relative to the new directory
                new_module_source = relpath(
                    module_source_absolute, self.target_directory_abs
                )
                content = content.replace(
                    f'"{module_source}"', f'"{new_module_source}"'
                )

            path.write_text(content)

    def _move_state_file_object(self):
        """Moves state file object inside S3 bucket"""
        state_bucket = self._find_auto_variable(
            self.source_directory_abs, "terraform_state_bucket"
        )
        source_key = self._build_state_file_s3_key(self.source_directory_abs)
        target_key = self._build_state_file_s3_key(self.target_directory_abs)

        try:
            response = self.s3_client.head_object(Bucket=state_bucket, Key=target_key)
        except ClientError as exc:
            if exc.response["Error"]["Code"] != "404":
                raise exc
        else:
            last_modified_date = response["LastModified"]
            raise RuntimeError(
                Colors.RED(
                    f"Target state file {target_key} already exists inside {state_bucket} bucket.\n"
                    "This means one of the following:\n"
                    " - the target directory has been applied in another branch;\n"
                    f" - the state file is a leftover and probably can be safely deleted"
                    f" (file was last modified on {last_modified_date.isoformat()})."
                )
            )

        try:
            self.s3_client.copy(
                CopySource={
                    "Bucket": state_bucket,
                    "Key": source_key,
                },
                Bucket=state_bucket,
                Key=target_key,
            )
        except ClientError as exc:
            if exc.response["Error"]["Code"] == "404":
                raise RuntimeError(
                    Colors.RED(
                        f"No state file has been found for {self.source_directory} "
                        f"directory inside {state_bucket} bucket.\n"
                        "Make sure that the directory is a valid terraform directory and has been applied."
                    )
                ) from exc
            raise exc from exc

    def _move_files(self, files: Iterable[Path]):
        """Moves terraform configuration files with git mv, skips subdirectories."""
        self.target_directory_abs.mkdir(parents=True, exist_ok=True)
        sources = [str(f) for f in files]
        self.repo.index.move([*sources, str(self.target_directory_abs)])

    def _verify_source_directory(self) -> List[Path]:
        """Verify that the source directory is a valid terraform directory, returns paths of files to be moved"""
        if not self.source_directory_abs.exists():
            raise RuntimeError(
                Colors.RED(
                    f"Source directory {self.source_directory_abs} does not exist."
                )
            )

        files_to_move = []
        has_terraform_files = False

        git_files = set(self.repo.git.ls_files().splitlines())
        rel_source_path = Path(
            *Path(calc_repo_path(self.source_directory_abs)).parts[1:]
        )

        for path in self.source_directory_abs.iterdir():
            if path.is_file() and str(rel_source_path / path.name) in git_files:

                if path.name.endswith(".tf"):
                    has_terraform_files = True

                files_to_move.append(path)

        if not has_terraform_files:
            raise RuntimeError(
                Colors.RED(
                    f"No terraform configuration files (*.tf) found in {calc_repo_path(self.source_directory_abs)}\n"
                    "If you wish to move a subdirectory, run this tool for its path."
                )
            )

        return sorted(files_to_move)

    def _verify_target_directory(self):
        """Verify that the target directory is empty"""
        self.target_directory_abs.mkdir(parents=True, exist_ok=True)
        if any(self.target_directory_abs.iterdir()):
            raise RuntimeError(
                Colors.RED(
                    f"Target directory {self.target_directory_abs} is not empty."
                )
            )

    def _verify_environment(self):
        error_message = Colors.RED(
            "Source and target directories must be in the same AWS account / environment"
        )

        source_state_bucket = self._find_auto_variable(
            self.source_directory_abs, "terraform_state_bucket"
        )
        target_state_bucket = self._find_auto_variable(
            self.target_directory_abs, "terraform_state_bucket"
        )

        if source_state_bucket != target_state_bucket:
            raise RuntimeError(error_message)

        source_environment = self._find_auto_variable(
            self.source_directory_abs, "environment", strict=False
        )
        target_environment = self._find_auto_variable(
            self.target_directory_abs, "environment", strict=False
        )

        if source_environment is None and target_environment is None:
            return

        if source_environment != target_environment:
            raise RuntimeError(error_message)

    def _diff_autovars(self):
        """Find .auto.fvars files that are available in scope of the source directory, but not in scope of the target"""
        source_var_files = self._find_variable_files(self.source_directory_abs.parent)
        target_var_files = self._find_variable_files(self.target_directory_abs.parent)

        missing_target_var_files = set(source_var_files) - set(target_var_files)
        if missing_target_var_files:
            print(
                Colors.YELLOW(
                    "Following .auto.tfvars files are not available in the scope of target directory:"
                )
            )
            for file in missing_target_var_files:
                file = Path(file)
                print("\t", Colors.YELLOW(f"{calc_repo_path(file.parent)}/{file.name}"))
            print(
                Colors.YELLOW(
                    "If your configuration depends on any of the variables defined in the above files, "
                    "make sure these variables are also defined in the target directory.\n"
                )
            )

    def run(self, skip_confirmation: bool = False):
        """Run"""

        if self.source_directory_abs == self.target_directory_abs:
            raise RuntimeError(Colors.RED("Source and target paths are the same."))

        files_to_move = self._verify_source_directory()
        self._verify_target_directory()
        self._verify_environment()

        state_bucket = self._find_auto_variable(
            self.source_directory_abs, "terraform_state_bucket"
        )
        source_state_file = self._build_state_file_s3_key(self.source_directory_abs)
        target_state_file = self._build_state_file_s3_key(self.target_directory_abs)

        print(
            "\nRemote state:\n"
            f"\tBucket:\t\t{Colors.BOLD(state_bucket)}\n"
            f"\tSource key: \t{Colors.BOLD(source_state_file)}\n"
            f"\tTarget key: \t{Colors.BOLD(target_state_file)}\n"
        )

        print(
            "Following files will be moved:\n"
            f"\t{Colors.BOLD(calc_repo_path(self.source_directory_abs))}"
        )
        for tf_file in files_to_move[:-1]:
            print(f"\t├────{Colors.BOLD(tf_file.name)}")
        print(f"\t└────{Colors.BOLD(files_to_move[-1].name)}")
        print(
            f"to the new directory: \n\t{Colors.BOLD(calc_repo_path(self.target_directory_abs))}\n"
        )

        self._diff_autovars()
        # fmt: off
        if (
            not skip_confirmation
            and input(Colors.BOLD(
                "This tool will move remote state file and local directory contents; proceed? (y/N) "
                )).lower() != "y"
        ):
            raise RuntimeError("Aborted.\n")
        # fmt: on

        self._adjust_modules_sources(
            file for file in files_to_move if file.name.endswith(".tf")
        )
        self._move_files(files_to_move)
        self._move_state_file_object()

        print(
            f"{Colors.GREEN('Remote state file moved successfuly.')}\n"
            f"{Colors.GREEN('Local directory contents moved successfuly.')}"
        )
        print(
            f"Adjust configuration in the target directory if needed.\n"
            f"Run {Colors.UNDERLINE('tf ' + str(self.target_directory) + ' init')}\n"
            f"and then {Colors.UNDERLINE('tf ' + str(self.target_directory) + ' plan')}\n"
            f"to make sure that there are no undesired infrastructure changes.\n"
            f"Remember to {Colors.UNDERLINE('commit local changes and create a PR')}.\n"
        )



"""setup.py controls the build, testing, and distribution of the egg"""
from __future__ import print_function

import re
import os.path

from setuptools import setup, find_packages


VERSION_REGEX = re.compile(
    r"""
    ^__version__\s=\s
    ['"](?P<version>.*?)['"]
""",
    re.MULTILINE | re.VERBOSE,
)

VERSION_FILE = os.path.join("terrawrap", "version.py")


def get_long_description():
    """Reads the long description from the README"""
    this_directory = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(this_directory, "README.md"), encoding="utf-8") as file:
        return file.read()


def get_version():
    """Reads the version from the package"""
    with open(VERSION_FILE, encoding="utf-8") as handle:
        lines = handle.read()
        result = VERSION_REGEX.search(lines)
        if result:
            return result.groupdict()["version"]
        raise ValueError("Unable to determine __version__")


def get_requirements():
    """Reads the installation requirements from requirements.txt"""
    with open("requirements.txt", encoding="utf-8") as reqfile:
        return [
            line
            for line in reqfile.read().split("\n")
            if not line.startswith(("#", "-"))
        ]


setup(
    name="terrawrap",
    python_requires=">=3.10.0",
    version=get_version(),
    description="Set of Python-based CLI tools for working with Terraform configurations",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    # Get strings from http://www.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.14",
    ],
    keywords="",
    author="Amplify Education",
    author_email="github@amplify.com",
    url="https://github.com/amplify-education/terrawrap",
    license="MIT",
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=get_requirements(),
    test_suite="nose.collector",
    scripts=[
        "bin/tf",
        "bin/tf_apply",
        "bin/backend_check",
        "bin/pipeline_check",
        "bin/plan_check",
        "bin/visualize",
        "bin/graph_apply",
        "bin/tf_move",
    ],
)



"""Utility for applying effects to terminal text with ANSI escape sequences."""
from enum import Enum
from typing import Any


class TerminalColors(Enum):
    """Utility for applying effects to terminal text with ANSI escape sequences."""

    RESET = "\033[0m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    RED = "\033[91m"
    BOLD = "\033[0;1m"
    UNDERLINE = "\033[0;4m"

    def __call__(self, text: Any):
        return self.value + str(text) + self.RESET.value  # type: ignore
